
from copy import deepcopy
from collections import OrderedDict
from collections import Counter
from collections import namedtuple
from itertools import islice

from .. util.filesystem import write_file
from .. util.filesystem import read_file
from .. util.filesystem import apply_file_extension
from .. util.iter import OrderedDefaultDict
from .. util.iter import iterator_to_list
from .. util.iter import base_class_names
from .. util.iter import iteration_depth
from .. util.iter import modify_iteration_depth
from .. util.iter import transpose
from .. util.iter import index_sequence
from .. util.iter import ordered_unique
from .. util.iter import is_vengeance_class
from .. util.iter import is_flux_row_class
from .. util.text import p_json_dumps

from . flux_row_cls import flux_row_cls


class flux_cls:

    def __init__(self, matrix=None):
        self._num_cols = None

        self.headers = OrderedDict()
        self.matrix  = self._to_flux_rows(matrix)

    @property
    def header_values(self):
        """
        (these may not necessarily be the same values as self.matrix[0].values)
        """
        return list(self.headers.keys())

    @property
    def num_rows(self):
        """ header row not included in count """
        return len(self.matrix) - 1

    @property
    def num_cols(self):
        """ maximum number of columns of all rows """
        if self._num_cols is None:
            m = [row.values for row in self.matrix]
            self._num_cols = max(map(len, m))

        return self._num_cols

    @property
    def is_jagged(self):
        num_cols = len(self.matrix[0].values)

        for row in self:
            if len(row.values) != num_cols:
                return True

        return False

    @property
    def is_empty(self):
        """ determine if matrix has any values """
        for row in self.matrix:
            if row.values:
                return False

        return True

    def _to_flux_rows(self, m):
        if m is None:
            return [flux_row_cls({}, [])]

        m = self.__validate_matrix(m)
        self.__validate_headers(m[0])

        if not self.headers:
            self.headers = index_sequence(m[0], as_strings=True)

        if self._num_cols is None:
            self._num_cols = max(map(len, m))

        headers = self.headers
        flux_m  = [flux_row_cls(headers, row) for row in m]

        return flux_m

    def _reapply_header_names(self, headers):
        """
        re-assigning self.headers will de-reference flux_row_cls._headers
        instead, self.headers.clear() must be called, followed by addition of items
        """
        if len(headers) != self._num_cols:
            self._num_cols = None

        self.headers.clear()
        for i, header in enumerate(headers):
            self.headers[header] = i

    def execute_commands(self, commands, profile=False):
        if profile:
            from line_profiler import LineProfiler
            profiler = LineProfiler()
        else:
            profiler = None

        commands = self.__parse_commands(commands)

        for command in commands:
            method = command.method
            if profiler:
                method = profiler(method)

            if command.num_args == 0:
                method()
            elif command.num_args == 1:
                method(command.args)
            else:
                method(*command.args)

        if profiler:
            profiler.print_stats()

    def __parse_commands(self, commands):
        parsed = []
        nt_cls = namedtuple('command_nt', ['name',
                                           'method',
                                           'args',
                                           'num_args'])

        for command in commands:
            if isinstance(command, (list, tuple)):
                name, args = command
                if isinstance(args, str):
                    num_args = 1
                else:
                    num_args = len(args)
            else:
                name = command
                args = None
                num_args = 0

            method = getattr(self, name)
            parsed.append(nt_cls(name, method, args, num_args))

        return parsed

    def matrix_by_headers(self, *headers):
        """
        performance enhancements:
            convert all headers to indices first
            take values on demand
            for c in range(indices):
                columns.append([row.values[c] for row in self.matrix])

            store columns that have already been used in dict
            self._num_cols does not need to be reset if length of headers
            equal to self._num_cols
        """
        headers = modify_iteration_depth(headers, depth=1)
        columns = transpose([row.values for row in self]) or [[]]

        m = []
        for header in headers:
            header = self.__renamed_or_inserted_header(header)

            column = [header]
            if header not in self.headers:
                column += [None] * self.num_rows
            else:
                i = self.headers[header]
                column += columns[i]

            m.append(column)

        m = transpose(m)

        self.headers.clear()
        self._num_cols = None

        self.matrix = self._to_flux_rows(m)

    def __renamed_or_inserted_header(self, header):
        """
        to ensure new columns are being created intentionally and not because
        of spelling errors etc, inserted headers must be surrounded by parenthesis
        """
        if not isinstance(header, (dict, str)):
            raise ValueError('{}\nheader must be either dictionary or string'.format(header))

        if isinstance(header, dict):                            # renamed: {'header_a': 'header_z'}
            h_old, h_new = tuple(header.items())[0]
            self.headers[h_new] = self.headers[h_old]

            return h_new

        if header.startswith('(') and header.endswith(')'):     # inserted: '(inserted_header)'
            return header[1:-1]

        if header not in self.headers:
            raise ValueError("'{header}' does not exist\ninserted columns should be surrounded "
                             "by parenthesis, ie '({header})' not '{header}'".format(header=header))

        return header

    def rename_columns(self, old_to_new_headers):
        """
        :param old_to_new_headers: a dictionary of {h_old: h_new} headers
        """
        renamed = self.matrix[0].values

        for h_old, h_new in old_to_new_headers.items():
            if h_old not in self.headers:
                raise KeyError('invalid header name {}'.format(h_old))

            i = self.headers[h_old]
            renamed[i] = h_new

        self._reapply_header_names(renamed)

    def insert_columns(self, *inserted):
        """
        :param inserted: (before, header_name)

        eg names:
            flux.insert_columns('header_a')                     insert column at index 0
            flux.insert_columns((3, 'header_a'))                insert column at index 3
            flux.insert_columns(('header_a', 'header_b'))       insert column at index 0
        """
        inserted = modify_iteration_depth(inserted, depth=0)
        if not inserted:
            return

        nd = iteration_depth(inserted)
        if nd == 0:
            inserted = [(0, inserted)]
        elif nd == 1:
            inserted = [inserted]

        names = [n[1] for n in inserted]
        names = self.__validate_column_names(names)

        invalid = list(self.headers.keys() & set(names))
        if invalid:
            raise ValueError('column name(s) conflict with existing headers:\n{}'.format(invalid))

        header_values = self.matrix[0].values
        for before, header in inserted:

            i = None
            if isinstance(before, int):
                i = before
            elif isinstance(before, str):
                if before == '*f':
                    i = 0
                elif before == '*l':
                    i = len(header_values)
                else:
                    try:
                        i = header_values.index(before)
                    except ValueError as e:
                        raise ValueError("header '{}' does not exist".format(before)) from e

            if i is None:
                raise KeyError("insertion header '{}' does not exist".format(before))

            header_values.insert(i, header)

        indices = []
        for _, header in inserted:
            indices.append(header_values.index(header))

        indices.sort()

        for i in indices:
            for row in self:
                row.values.insert(i, None)

        self._reapply_header_names(header_values)

    def append_columns(self, *names):
        names = self.__validate_column_names(names)

        invalid = list(self.headers.keys() & set(names))
        if invalid:
            raise ValueError('column name(s) conflict with existing headers:\n{}'.format(invalid))

        self.matrix[0].values.extend(names)
        for header in names:
            self.headers[header] = len(self.headers)

        nones = [None for _ in range(len(names))]
        for row in self:
            row.values.extend(nones)

        self._num_cols = None

    def delete_columns(self, *names):
        """ if negative indices? """
        names = self.__validate_column_names(names)
        indices = [self.headers.get(h, h) for h in names]

        invalid = [i for i in indices if not isinstance(i, int)]
        if invalid:
            raise ValueError('column(s) do not exist:\n{}'.format(invalid))

        invalid = [i for i in indices if i < 0]
        if invalid:
            raise ValueError('negative indices not supported')

        indices.sort(reverse=True)

        for row in self.matrix:
            for c in indices:
                del row.values[c]

        self._reapply_header_names(self.matrix[0].values)

    def fill_jagged_columns(self):
        """ :returns indices of rows whose columns were extended """

        max_cols = self.num_cols

        row_indices = []
        for i, row in enumerate(self.matrix):
            values = row.values
            num_missing = max_cols - len(values)

            if num_missing > 0:
                row_indices.append(i)
                nones = [None for _ in range(num_missing)]
                values.extend(nones)

        return row_indices

    def insert_rows(self, i, rows):
        if self.is_empty:
            self.matrix = self._to_flux_rows(rows)
            return

        if i == 0:
            self.headers.clear()
            del self.matrix[0]
        elif is_vengeance_class(rows):
            rows = list(rows.rows())[1:]

        self.matrix[i:i] = self._to_flux_rows(rows)

        self._num_cols = None

    def append_rows(self, rows):
        if self.is_empty:
            self.matrix = self._to_flux_rows(rows)
            return

        if is_vengeance_class(rows):
            rows = list(rows.rows())[1:]

        self.matrix.extend(self._to_flux_rows(rows))

        self._num_cols = None

    def sort(self, *f, reverse=None):
        """ in-place sort
        sort priority of columns proceeds from left to right

        eg:
            flux.sort('col_a')

        eg multiple columns:
            flux.sort('col_a', 'col_b', 'col_c',
                      reverse=[True, True, True])
        """
        self.matrix[1:] = self.__sort_rows(f, reverse)

    def sorted(self, *f, reverse=None):
        """ returns new flux after sorting
        sort priority of columns proceeds from left to right

        eg:
            flux = flux.sorted('col_a')

        eg multiple columns:
            flux = flux.sorted('col_a', 'col_b', 'col_c',
                               reverse=[True, True, True])
        """
        m = [self.matrix[0].values.copy()]
        m.extend([row.values.copy() for row in self.__sort_rows(f, reverse)])

        flux = self.__class__()
        flux._num_cols = self._num_cols
        flux.matrix = flux._to_flux_rows(m)

        return flux

    def __sort_rows(self, f, order):
        """
        since sort priority of columns proceeds from left to right, (the first
        column specified is the last one sorted), sort functions and sort orders
        must be reversed
        """
        if isinstance(order, bool):
            order = [order]
        else:
            order = order or []

        if not isinstance(order, (list, tuple)):
            raise TypeError('reverse must be either a list of boolean')

        f = modify_iteration_depth(f, 1)

        for _ in range(len(f) - len(order)):
            order.append(False)

        o = reversed(order)
        f = reversed(f)

        m = self.matrix[1:]
        for _f_, _o_ in zip(f, o):
            _f_ = self.__row_values_accessor(_f_)
            m.sort(key=_f_, reverse=_o_)

        return m

    def filter(self, f):
        """ in-place """
        self.matrix[1:] = [row for row in self if f(row)]

    def filtered(self, f):
        """ return new flux """
        m = [self.matrix[0].values.copy()]
        m.extend([row.values.copy() for row in self if f(row)])

        flux = self.__class__()
        flux._num_cols = self._num_cols
        flux.matrix = flux._to_flux_rows(m)

        return flux

    def filter_by_unique(self, *f):
        self.__filter_unique_rows(f, in_place=True)

    def filtered_by_unique(self, *f):
        return self.__filter_unique_rows(f, in_place=False)

    def __filter_unique_rows(self, f, in_place):
        uniq = set()

        def evaluate_unique(row):
            v = f(row)
            if v not in uniq:
                uniq.add(v)
                return True
            else:
                return False

        f = self.__row_values_accessor(f)
        if in_place:
            self.filter(evaluate_unique)
        else:
            return self.filtered(evaluate_unique)

    def unique_values(self, *f):
        """ :return: list of unique values within column(s), original order is preserved """
        f = self.__row_values_accessor(f)
        return ordered_unique(f(row) for row in self)

    def contiguous_rows(self, *f):
        """ :return: list of rows where values are contiguous """
        rows = []
        for i_1, i_2 in self.contiguous_indices(f):
            rows.append(self.matrix[i_1:i_2])

        return rows

    def contiguous_indices(self, *f):
        """ :return: list of (i_1, i_2) row indices where values are contiguous
        eg:
            for i_1, i_2 in flux.contiguous_indices('col_a', 'col_b'):
                rows = flux[i_1:i_2]
        """
        if self.num_rows == 0:
            return []

        f = self.__row_values_accessor(f)

        rows = iter(self)
        v_p  = f(next(rows))

        indices = [1]
        for i, row in enumerate(rows, 2):
            v = f(row)
            if v != v_p:
                indices.append(i)
                v_p = v

        indices.append(len(self.matrix))
        indices = [(i_1, i_2) for i_1, i_2 in zip(indices, indices[1:])]

        return indices

    def replace_matrix(self, m):
        self._num_cols = None
        self.headers   = None

        self.matrix = self._to_flux_rows(m)

    def index_row(self, *f):
        """ dictionary of {f(row): row}
        overwrites row values for non-unique key values
        """
        f = self.__row_values_accessor(f)
        return OrderedDict((f(row), row) for row in self)

    def index_rows(self, *f):
        """ dictionary of {f(row): [rows]}
        preserves rows in a list for non-unique key values
        """
        f = self.__row_values_accessor(f)
        items = [(f(row), row) for row in self]

        return OrderedDefaultDict(list, items)

    def namedtuples(self):
        try:
            nt_cls = namedtuple('flux_row_nt', self.headers.keys())
            return [nt_cls(*row.values) for row in self]
        except ValueError as e:
            import re

            names = [n for n in self.headers.keys() if re.search('^[^a-z]|[ ]', str(n), re.I)]
            if names:
                raise ValueError('invalid field(s) for namedtuple constructor: {}'.format(names)) from e
            else:
                raise e

    def enumerate_rows(self, start=0):
        """ to assist with debugging """
        if 'i' in self.headers:
            raise AssertionError("column 'i' already exists as a header name")

        for i, row in enumerate(self.matrix, start):
            row.__dict__['i'] = i

    def copy(self, deep=False):
        if deep:
            return deepcopy(self)

        m = [row.values.copy() for row in self.matrix]

        flux = self.__class__()
        flux._num_cols = self._num_cols
        flux.matrix = flux._to_flux_rows(m)

        return flux

    def to_csv(self, path, encoding=None):
        m = list(self.rows())

        path = apply_file_extension(path, '.csv')
        write_file(path, m, encoding=encoding)

    @classmethod
    def from_csv(cls, path, encoding=None):
        path = apply_file_extension(path, '.csv')
        m = read_file(path, encoding=encoding)

        return cls(m)

    def to_json(self, path=None, encoding=None):
        j = [row.dict() for row in self]
        j = p_json_dumps(j)

        if path is None:
            return j

        path = apply_file_extension(path, '.json')
        write_file(path, j, encoding=encoding)

    @classmethod
    def from_json(cls, path, encoding=None):
        path = apply_file_extension(path, '.json')

        rows = read_file(path, encoding=encoding)
        if rows:
            m = [list(d.values()) for d in rows]
            m.insert(0, list(rows[0].keys()))
        else:
            m = [[]]

        return cls(m)

    def serialize(self, path):
        """
        while convenient, using pickle also introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        path = apply_file_extension(path, '.flux')
        write_file(path, self)

    @classmethod
    def deserialize(cls, path):
        """
        while convenient, using pickle also introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        path = apply_file_extension(path, '.flux')
        return read_file(path)

    def rows(self, r_1='*h', r_2='*l'):
        r_1 = self._matrix_index(r_1)
        r_2 = self._matrix_index(r_2)

        return (row.values for row in islice(self.matrix, r_1, r_2 + 1))

    def flux_rows(self, r_1='*h', r_2='*l'):
        r_1 = self._matrix_index(r_1)
        r_2 = self._matrix_index(r_2)

        return iter(islice(self.matrix, r_1, r_2 + 1))

    def _matrix_index(self, reference):
        if isinstance(reference, (int, float)):
            return int(reference)

        anchors = {'*h': 0,
                   '*f': 1,
                   '*l': len(self.matrix) - 1}

        r_i = anchors.get(reference)
        if r_i is None:
            raise KeyError("invalid index '{}'".format(reference))

        return r_i

    def __row_values_accessor(self, f):
        """ convert f into a function that can retrieve values for each row in self.matrix """

        def row_values(row):
            """ return muliple values (as tuple) from each row """
            return tuple(row.values[c] for c in columns)

        def row_value(row):
            """ return single value from each row """
            return row.values[i]

        f = modify_iteration_depth(f, depth=0)
        self.__validate_accessor(f)

        if callable(f):
            return f

        if isinstance(f, (list, tuple)):
            columns = tuple(self.headers.get(n, n) for n in f)
            f = row_values
        elif isinstance(f, str):
            i = self.headers[f]
            f = row_value
        elif isinstance(f, int):
            i = f
            f = row_value

        return f

    def __validate_accessor(self, f):
        num_cols = len(self.headers) - 1

        if isinstance(f, (str, list, tuple)):
            _f_ = modify_iteration_depth(f, 1)

            invalid = [h for h in _f_
                         if isinstance(h, str)
                         if h not in self.headers]
            if invalid:
                raise KeyError("invalid column reference: {}".format(invalid))

            invalid = [i for i in _f_
                         if isinstance(i, int)
                         if i > num_cols]
            if invalid:
                raise KeyError("integer(s) out of bounds: {}".format(invalid))

        elif isinstance(f, int):
            if f > num_cols:
                raise KeyError("integer out of bounds: {}".format(f))

        else:
            raise TypeError("unhandled type conversion for '{}'".format(f))

    @staticmethod
    def __validate_matrix(m):
        base_names = set(base_class_names(m))

        if 'DataFrame' in base_names:
            raise NotImplementedError("conversion of 'DataFrame' not supported yet")

        if 'ndarray' in base_names:
            raise NotImplementedError("conversion of 'ndarray' not supported yet")

        if base_names & {'flux_cls', 'excel_levity_cls'}:
            m = list(m.rows())

        m = iterator_to_list(m)

        if iteration_depth(m) != 2:
            raise IndexError('matrix must have exactly 2 dimensions')

        if is_flux_row_class(m[0]):
            m = [m[0].names] + [row.values for row in m]

        return m

    @staticmethod
    def __validate_headers(headers):
        reserved = set(headers) & flux_row_cls.class_names
        if reserved:
            raise NameError('conflicting name(s) {} found in header row: {}'.format(list(reserved), headers))

    @staticmethod
    def __validate_column_names(names):
        names = modify_iteration_depth(names, depth=1)

        duplicates = [n for n, count in Counter(names).items() if count > 1]
        if duplicates:
            raise ValueError('duplicate column name(s) detected:\n{}'.format(duplicates))

        return names

    def __len__(self):
        return len(self.matrix)

    def __getitem__(self, row_or_col):
        """ returns flux_row, flux_rows slice, or column values
        eg:
            flux_row  = flux[3]
            flux_rows = flux[3:5]
            values    = flux['col']
        """
        # flux_row
        if isinstance(row_or_col, int):
            return self.matrix[row_or_col]

        # flux_rows (slice)
        if isinstance(row_or_col, slice):
            i_1, i_2, step = row_or_col.start, row_or_col.stop, row_or_col.step
            if i_1 is not None:
                if i_1 < 0:
                    i_1 = len(self.matrix) + i_1

            if i_2 is not None:
                if i_2 < 0:
                    i_2 = len(self.matrix) + i_2

            return [row for row in islice(self.matrix, i_1, i_2, step)]

        # column values
        if row_or_col in self.headers:
            i = self.headers[row_or_col]
            return [row.values[i] for row in self]

        raise KeyError("undefined reference: '{}'".format(row_or_col))

    def __setitem__(self, col_name, col_values):
        """ sets values to a single column

        number of values must be equal to length of self.matrix - 1
        eg:
            flux['col'] = [None] * flux.num_rows
        """
        col_values = modify_iteration_depth(col_values, 1)

        nd = iteration_depth(col_values)
        if nd == 0:
            col_values = [col_values]
        elif nd == 2:
            col_values = transpose(col_values)[0]

        if len(col_values) != self.num_rows:
            raise ValueError('number of rows in column must match the number of rows in flux')

        if col_name not in self.headers:
            # potentially inefficient to init with Nones when values are known...
            self.append_columns(col_name)

        i = self.headers[col_name]
        for row, v in zip(self, col_values):
            row.values[i] = v

    def __iter__(self):
        return islice(self.matrix, 1, None)

    def __add__(self, flux_b):
        flux_final = self.copy()
        flux_final.append_rows(flux_b)

        return flux_final

    def __repr__(self):
        if self.is_empty:
            return '(empty)'

        headers = list(self.headers.keys())
        headers = str(headers).replace("'", '')

        return '({:,})  {}'.format(self.num_rows, headers)






import _pickle as cpickle

import csv
import json
import os
import shutil

from datetime import datetime
from glob import glob

from .text import p_json_dumps
from .text import repr_


def read_file(path, encoding=None, mode='r'):
    """
    assumes text file unless extension is in these specialized formats:
        .csv
        .json
        .flux
        .pkl
        .pickle
    """
    assert_path_exists(path)
    extn = file_extension(path, include_dot=True)

    if extn.startswith('.xls') or extn in {'.7z', '.gzip', '.hd5'}:
        raise NotImplementedError

    if extn == '.csv':
        with open(path, mode, encoding=encoding) as f:
            return list(csv.reader(f))

    if extn == '.json':
        with open(path, mode, encoding=encoding) as f:
            return json.load(f)

    if extn in {'.flux', '.pkl', '.pickle'}:
        if not mode.endswith('b'):
            mode += 'b'
        with open(path, mode) as f:
            return cpickle.load(f)

    with open(path, mode, encoding=encoding) as f:
        return f.read()


def write_file(path, data, encoding=None, mode='w'):
    """
    assumes text file unless extension is in these specialized formats:
        .csv
        .json
        .flux
        .pkl
        .pickle
    """
    extn = file_extension(path, include_dot=True)

    if extn.startswith('.xls') or extn in {'.7z', '.gzip', '.hd5'}:
        raise NotImplementedError

    if extn == '.csv':
        with open(path, mode, encoding=encoding) as f:
            csv.writer(f, lineterminator='\n').writerows(data)
        return

    if extn == '.json':
        if not isinstance(data, str):
            data = p_json_dumps(data, ensure_ascii=(encoding is None))

        with open(path, mode, encoding=encoding) as f:
            f.write(data)

        return

    if extn in {'.flux', '.pkl', '.pickle'}:
        if not mode.endswith('b'):
            mode += 'b'

        with open(path, mode) as f:
            cpickle.dump(data, f)

        return

    if not isinstance(data, str):
        # convert data to string: f.write() is much faster than f.writelines()
        data = repr_(data, concat='\n', quotes=False, wrap=False)

    with open(path, mode, encoding=encoding) as f:
        f.write(data)


def clear_dir(f_dir, allow_not_exist=False):
    f_dir = standardize_dir(f_dir)

    if not os.path.exists(f_dir):
        if allow_not_exist is True:
            return
        else:
            raise FileNotFoundError("cannot clear: '{}', directory does not exist".format(f_dir))

    for file_content in os.listdir(f_dir):
        path = f_dir + file_content

        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def copy_dir(s_dir, d_dir, exclude_dirs=None):
    exclude_dirs = exclude_dirs or set()

    s_dir = standardize_dir(s_dir)
    d_dir = standardize_dir(d_dir)

    if not os.path.exists(d_dir):
        os.makedirs(d_dir)

    for file_content in os.listdir(s_dir):
        if file_content in exclude_dirs:
            continue

        s_path = s_dir + file_content
        d_path = d_dir + file_content

        if os.path.isdir(s_path):
            shutil.copytree(src=s_path, dst=d_path)
        else:
            shutil.copy(src=s_path, dst=d_path)


def file_creation_date(path):
    unix_t = os.path.getctime(path)
    return datetime.fromtimestamp(unix_t)


def file_last_modified(path):
    unix_t = os.stat(path).st_mtime
    return datetime.fromtimestamp(unix_t)


def standardize_dir(f_dir, pathsep='/'):
    """ lowercase and make sure directory follows predictable pattern
    pathsep=os.path.sep?
    """
    f_dir = (f_dir.replace('\\', pathsep)
                  .replace('/', pathsep)
                  .lower()
                  .strip())

    if not f_dir.endswith(pathsep):
        f_dir += pathsep

    return f_dir


def standardize_file_name(f_name):
    return f_name.lower().strip()


def standardize_path(path):
    f_dir, f_name = parse_path(path)

    f_dir  = standardize_dir(f_dir)
    f_name = standardize_file_name(f_name)

    return f_dir + f_name


def sanatize_file_name(f_name):
    """
    replace illegal Windows file name characters with '-'

    (there's an additional set of path characters that are
     illegal for Windows' built-in compression)
    """
    invalid_chrs = {'\\': '-',
                    '/':  '-',
                    ':':  '-',
                    '*':  '-',
                    '?':  '-',
                    '<':  '-',
                    '>':  '-',
                    '|':  '-',
                    '"':  '-'}

    for k, v in invalid_chrs.items():
        f_name = f_name.replace(k, v)

    f_name = f_name.strip()
    if not f_name:
        raise IOError('file name is blank')

    return f_name


def assert_path_exists(path):
    f_dir, f_name = parse_path(path)

    if not os.path.exists(f_dir):
        raise FileNotFoundError("invalid directory: '{}'".format(f_dir))

    if not os.path.exists(path):
        extn  = file_extension(f_name, include_dot=True)
        retry = f_name.replace(extn, '.*')
        path  = glob(standardize_dir(f_dir) + retry)

        if path:
            retry_name = path[0].split('\\')[-1]
            retry_extn = file_extension(retry_name, include_dot=True)
            msg = "invalid file extension: '{}' (did you mean '{}' ?)".format(extn, retry_extn)
        else:
            msg = "'{}' not found within directory '{}'".format(f_name, f_dir)

        raise FileNotFoundError(msg)


def parse_path(path):

    if os.path.isdir(path):
        f_dir  = path
        f_name = ''
    else:
        f_dir, f_name = os.path.split(path)

    return f_dir, f_name


def file_extension(f_name, include_dot=True):
    _, extn = os.path.splitext(f_name)
    if include_dot is False:
        extn = extn.replace('.', '')

    extn = extn.lower().strip()

    return extn


def apply_file_extension(f_name, extn):
    if not extn.startswith('.'):
        extn = '.' + extn

    if f_name.endswith(extn):
        return f_name

    return f_name + extn






from collections import OrderedDict
from collections import namedtuple


class flux_row_cls:

    class_names = {'_headers',
                   'values',
                   'names',
                   'dict',
                   'namedtuples'}

    def __init__(self, headers, values):
        """
        :param headers: OrderedDict of {'header': index}

            * headers is a single dictionary passed byref from the
              flux_cls to many flux_row_cls instances
            * this eliminates need for all flux_row_cls objects
              to maintain a seprate copy of these mappings and
              allows for centralized and instantaneous updatdes

        :param values: list of underlying data

        self.__dict__ is used to set attributes in __init__ so as to avoid
        premature __setattr__ lookups
        """
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values

    @property
    def names(self):
        return list(self._headers.keys())

    @property
    def view_as_array(self):
        """ for development purposes; used purely to trigger a debugging feature in PyCharm

        PyCharm will recognize returned the (name, value) pairs as an ndarray
        and enable the "...view as array" option in the debugger which displays
        values in a special window
        """
        import numpy
        return numpy.transpose([self.names, self.values])

    def dict(self):
        return OrderedDict(zip(self.names, self.values))

    def namedtuples(self, nt_cls=None):
        try:
            if nt_cls is None:
                nt_cls = namedtuple('flux_row_nt', self.names)

            return nt_cls(*self.values)
        except ValueError as e:
            import re

            names = [n for n in self.names if re.search('^[^a-z]|[ ]', str(n), re.I)]
            if names:
                raise ValueError('invalid field(s) for namedtuple constructor: {}'.format(names)) from e
            else:
                raise e

    def __getattr__(self, name):
        """  eg:
             v = row.column
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __getitem__(self, name):
        """ eg:
            v = row['column']
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __setattr__(self, name, value):
        """ eg:
            row.column = v
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __setitem__(self, name, value):
        """ eg:
            row['column'] = v
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError) as e:
            raise AttributeError(self.__attr_err_msg(name)) from e

    def __attr_err_msg(self, name):
        if isinstance(name, slice):
            return 'slice should be used on row.values\n(eg, row.values[2:5], not row[2:5])'

        names = '\n\t'.join(self.names)
        return "No flux_row_cls column named '{}'\navailable columns:\n\t{}".format(name, names)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        super().__setattr__('__dict__', d)

    def __len__(self):
        return len(self.values)

    def __bool__(self):
        return bool(self.values)

    def __iter__(self):
        return iter(self.values)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self._headers) + hash(tuple(self.values))

    def __repr__(self):
        i = self.__dict__.get('i', '')
        if i != '':
            i = '({:,})  '.format(i)

        return '{}{}'.format(i, repr(self.values))


class lev_row_cls(flux_row_cls):

    class_names = {'_headers',
                   'values',
                   'names',
                   'dict',
                   'namedtuples',
                   'address'}

    def __init__(self, headers, values, address):
        super().__init__(headers, values)
        self.__dict__['address'] = address

    def __repr__(self):
        return '{} {}'.format(self.address, repr(self.values))





from . dates import to_datetime
from . dates import attempt_to_datetime
from . dates import is_valid_date

from . filesystem import read_file
from . filesystem import write_file

from . iter import transpose
from . text import print_runtime


__all__ = ['to_datetime',
           'attempt_to_datetime',
           'is_valid_date',
           'read_file',
           'write_file',
           'transpose',
           'print_runtime']




import os

from vengeance import open_workbook
from vengeance import close_workbook
from vengeance import get_worksheet
from vengeance import flux_cls
from vengeance import excel_levity_cls

wb = None
__levs = {}

project_dir = os.getcwd() + '\\files\\'


def open_project_workbook(open_new_instance=True,
                          read_only=False,
                          update_links=True):
    global wb

    if wb is not None:
        return

    if open_new_instance:
        excel_app = None
    else:
        excel_app = 'any'

    wb = open_workbook(project_dir + 'example.xlsm',
                       excel_app,
                       read_only=read_only,
                       update_links=update_links)


def close_project_workbook(save=True):
    global wb

    if wb is None:
        return

    close_workbook(wb, save)
    wb = None


def worksheet_to_lev(ws,
                     meta_r=1,
                     header_r=2,
                     c_1=None,
                     c_2=None):

    global __levs

    if isinstance(ws, str):
        ws_name = ws.lower()
    elif isinstance(ws, excel_levity_cls):
        ws = ws.worksheet
        ws_name = ws.Name
    elif ws.__class__.__name__ in {'CDispatch', '_Worksheet'}:
        ws = ws
        ws_name = ws.Name
    else:
        ws_name = ws

    if ws_name in {'sheet1', 'empty sheet'}:
        header_r = 1
        meta_r   = 0
    else:
        c_1 = c_1 or 'B'

    k = (ws_name,
         meta_r,
         header_r,
         c_1,
         c_2)

    if k in __levs:
        return __levs[k]

    ws = get_worksheet(wb, ws)

    headers = {}
    if (c_1 or c_2) and (header_r > 0):
        headers.update(excel_levity_cls.index_headers(ws, header_r))
    if (c_1 or c_2) and (meta_r > 0):
        headers.update(excel_levity_cls.index_headers(ws, meta_r))

    c_1 = headers.get(c_1, c_1)
    c_2 = headers.get(c_2, c_2)

    lev = excel_levity_cls(ws,
                           meta_r=meta_r,
                           header_r=header_r,
                           first_c=c_1,
                           last_c=c_2)
    __levs[k] = lev

    return lev


def worksheet_to_flux(ws,
                      meta_r=1,
                      header_r=2,
                      c_1=None,
                      c_2=None):

    lev = worksheet_to_lev(ws,
                           meta_r,
                           header_r,
                           c_1,
                           c_2)
    return flux_cls(lev)


def write_to_worksheet(ws,
                       m,
                       r_1='*h',
                       c_1='B',
                       c_2=None):

    lev = worksheet_to_lev(ws, c_1=c_1, c_2=c_2)
    lev.activate()

    was_filtered = lev.has_filter

    if r_1 != '*a':
        lev.clear('*f %s:*l *l' % r_1)

    lev['*f %s' % r_1] = m

    if was_filtered:
        lev.reapply_filter()











from math import ceil

from collections import OrderedDict
from collections import defaultdict
from collections import Iterable


def modify_iteration_depth(v, depth=0):
    """
    eg:
        'a'               = modify_iteration_depth([['a']], depth=0)
        ['a', 'b', 'c']   = modify_iteration_depth([['a', 'b', 'c']], depth=0)
        [['a']]           = modify_iteration_depth('a', depth=2)
        [['a', 'b', 'c']] = modify_iteration_depth(['a', 'b', 'c'], depth=2)

    (but only evaluates depth of first element of iterable)
    eg:
        ['a', [['b']], [['c']]] = modify_iteration_depth(['a', [['b']], [['c']]], depth=0)
    """
    v  = iterator_to_list(v)
    nd = iteration_depth(v)

    # remove unneccesary nesting
    if nd > depth:
        for _ in range(nd - depth):
            if _is_subscriptable(v) and len(v) == 1:
                v = v[0]

    # apply additional nesting
    elif nd < depth:
        for _ in range(depth - nd):
            v = [v]

    return v


def iterator_to_list(v, recurse=False):
    if not _is_iterable(v):
        return v

    if is_vengeance_class(v):
        v = list(v.rows())
    elif not isinstance(v, (list, tuple)):
        v = list(v)

    if recurse:
        for i, _v_ in enumerate(v):
            v[i] = iterator_to_list(_v_, recurse=True)

    return v


def iteration_depth(v):
    """ determine number of nested iteration levels (number of dimensions)

    eg:
        0 = iteration_depth('abc')
        1 = iteration_depth(['abc'])
        1 = iteration_depth([])
        2 = iteration_depth([['abc'], ['bcd'] ])
        2 = iteration_depth([[]])

    (only evaluates the first element if iteration_depth > 1)
    eg:
        2 = iteration_depth([['abc'], 1, 2])
    """
    if _is_exhaustable_iterator(v):
        raise TypeError('cannot evaluate iteration depth of an exhaustable iterator')

    if isinstance(v, str):
        return 0

    if not isinstance(v, Iterable):
        return 0

    if not v:
        return 1

    if not isinstance(v, (list, tuple)):
        v = list(v)

    try:
        _v_ = v[0]
    except TypeError:
        return 1

    return 1 + iteration_depth(_v_)


def stride_sequence(sequence, stride_len):
    sequence = iterator_to_list(sequence)

    for i_1 in range(0, len(sequence), stride_len):
        i_2 = i_1 + stride_len
        yield sequence[i_1:i_2]


def divide_sequence(sequence, num_divisions):
    sequence = iterator_to_list(sequence)

    num_items  = len(sequence)
    stride_len = max(1, ceil(num_items / num_divisions))

    for i_1 in range(0, num_items, stride_len):
        i_2 = i_1 + stride_len
        yield sequence[i_1:i_2]


def is_vengeance_class(o):
    bases = set(base_class_names(o))
    if 'flux_cls' in bases or 'excel_levity_cls' in bases:
        return True

    return False


def is_flux_row_class(o):
    bases = set(base_class_names(o))
    return 'flux_row_cls' in bases


def base_class_names(o):
    return [b.__name__ for b in o.__class__.mro()]


def transpose(m):
    if iteration_depth(m) == 1:
        t = [[row] for row in m]
    else:
        t = [list(row) for row in zip(*m)]

    return t


def append_matrices(direction, *matrices, has_header=True):
    d = direction

    if d.startswith('v') or d.startswith('row'):
        return append_vertical(*matrices, has_header=has_header)

    if d.startswith('h') or d.startswith('col'):
        return append_horizontal(*matrices)

    raise ValueError("invalid direction: '{}'".format(direction))


def append_vertical(*matrices, has_header=True):

    def append(m_1, m_2):
        m_1 = iterator_to_list(m_1)
        m_2 = iterator_to_list(m_2)

        if len(m_1[0]) != len(m_2[0]):
            raise IndexError('vertical append requires matrices to have equal number of columns')

        if is_empty(m_1):
            return m_2

        if is_empty(m_2):
            return m_1

        if has_header:
            m_2 = m_2[1:]

        return m_1 + m_2

    m_final, matrices = matrices[0], matrices[1:]
    for m in matrices:
        m_final = append(m_final, m)

    return m_final


def append_horizontal(*matrices):
    """
    this is god-awfully slow

    add all columns together in single iteration
    zip(*matrices)?
    but must assert all equal num columns
    """

    def append(m_1, m_2):
        m_1 = iterator_to_list(m_1)
        m_2 = iterator_to_list(m_2)

        if len(m_1) != len(m_2):
            raise IndexError('horizontal append requires matrices to have equal number of rows')

        if is_empty(m_1):
            return m_2

        if is_empty(m_2):
            return m_1

        return [row_1 + row_2 for row_1, row_2 in zip(m_1, m_2)]

    m_final, matrices = matrices[0], matrices[1:]
    for m in matrices:
        m_final = append(m_final, m)

    return m_final


def is_empty(v):
    """ determine if iterable is composed entirely of empty iterables, eg, [] or [[]]
    eg:
        True  = is_empty( [] )
        True  = is_empty( [[]] )
        False = is_empty( [[None]] )
        False = is_empty( [[], [], []] )
    """
    m = iterator_to_list(v)
    num_rows = len(m)
    num_cols = len(m[0])

    return (num_rows == 1) and (num_cols == 0)


def index_sequence(sequence, start=0, as_strings=True):
    indices   = OrderedDict()
    nonunique = defaultdict(int)

    for i, v in enumerate(sequence, start):
        if v in {'', None}:
            v = '_None_'

        if v in nonunique:
            _v_ = '{} ({})'.format(v, nonunique[v] + 1)
        elif as_strings:
            _v_ = str(v)
        else:
            _v_ = v

        indices[_v_]  = i
        nonunique[v] += 1

    return indices


def ordered_unique(sequence):
    d = OrderedDict((k, None) for k in sequence)
    return list(d.keys())


def invert_mapping(d):
    """
    return new mapping of
    {value: key}
        from
    {key: value}

    eg:
        {'a': [-1, -1, -1]
         'b': 2
         'c': 3
         'd': 3}
                to
        {-1: ['a', 'a', 'a']
         2:  b                      # single items not in list
         3:  ['c', 'd']}
    """
    if not isinstance(d, dict):
        raise TypeError('value must be a dictionary')

    inverted = OrderedDefaultDict(list)
    for k, v in d.items():
        _k_, _v_ = v, k

        for _k_ in modify_iteration_depth(v, 1):
            inverted[_k_].append(_v_)

    inverted = OrderedDict(inverted)
    for k, v in inverted.items():
        if len(v) == 1:
            inverted[k] = v[0]

    return inverted


def _is_iterable(v):
    """ determine if value is an iteraterable data structure
    (strings are not considered to be iterable data structures in this function)

    eg:
        False = is_iterable('mike')
        True  = is_iterable(['m', 'i' 'k' 'e'])
    """
    return isinstance(v, Iterable) and not isinstance(v, str)


def _is_subscriptable(v):
    return isinstance(v, (list, tuple))


def _is_exhaustable_iterator(v):
    return hasattr(v, '__next__')


class OrderedDefaultDict(OrderedDict):

    def __init__(self, default_factory=None, *args, **kwargs):
        if default_factory and (not callable(default_factory)):
            raise TypeError('first argument must be callable')

        self.default_factory = default_factory

        if args and (self.default_factory is list):
            self.__append_values(args[0])
        else:
            OrderedDict.__init__(self, *args, **kwargs)

    def __append_values(self, sequence):
        """
        append all values as list if constructed default_factory is list
        """

        if isinstance(sequence, dict):
            sequence = OrderedDict(sequence.items()).items()
        elif not isinstance(sequence, (list, tuple)):
            raise ValueError

        for item in sequence:
            if len(item) == 2:
                self[item[0]].append(item[1])
            elif len(item) < 2:
                raise ValueError('need more than 1 value to unpack')
            elif len(item) > 2:
                raise ValueError('too many values to unpack (expected 2)')

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = value = self.default_factory()

        return value

    def defaultdict(self):
        return defaultdict(self.default_factory, self.items())

    def ordereddict(self):
        return OrderedDict(self)

    def count_values(self):
        d = OrderedDict()

        for k, v in self.items():
            if _is_iterable(v):
                d[k] = len(v)
            else:
                d[k] = 1

        return d

    def modify_iteration_depth(self, depth=0):
        for k, v in self.items():
            self[k] = modify_iteration_depth(v, depth)






import os
import sys
from textwrap import dedent

from logging import Logger
from logging import Formatter
from logging import FileHandler
from logging import StreamHandler
from logging import DEBUG

from concurrent.futures import ProcessPoolExecutor


class log_cls(Logger):

    def __init__(self, name, f_dir=None, log_format=None):
        super().__init__(name)

        self.log_format  = log_format
        self.formatter   = None
        self.child_desig = None
        self._callback   = None

        self._file_handler   = None
        self._stream_handler = None

        self.err_msg = ''

        self._set_level()
        self._add_formatter()
        self._add_file_handler(f_dir)
        self._add_stream_handler()

    def print_message(self, msg):
        fh = self._file_handler
        sh = self._stream_handler

        if fh:
            fh.formatter = None
        if sh:
            sh.formatter = None

        self.info(msg)

        if fh:
            fh.formatter = self.formatter
        if sh:
            sh.formatter = self.formatter

    def add_parent(self, p_log):
        self.parent = p_log
        self._close_stream_handlers()

    def add_callback_function(self, f):
        self._callback = f

    def callback(self):
        if self._callback:
            self._callback()

    def _set_level(self):
        self.setLevel(DEBUG)

    def _add_formatter(self):
        if self.log_format is None:
            self.log_format = '%(asctime)s - %(levelname)s - %(message)s'

        self.formatter = Formatter(self.log_format)
        self.formatter.default_time_format = '%Y-%m-%d %I:%M:%S %p'

    def _add_file_handler(self, f_dir):
        if f_dir is None:
            return

        if not os.path.exists(f_dir):
            os.makedirs(f_dir)

        f_name = self.name
        if not f_name.endswith('.log'):
            f_name += '.log'

        h = FileHandler(str(f_dir) + f_name, mode='w')
        h.setLevel(self.level)
        h.setFormatter(self.formatter)

        self.addHandler(h)
        self._file_handler = h

    def _add_stream_handler(self):
        h = StreamHandler()
        h.setLevel(self.level)
        h.setFormatter(self.formatter)

        self.addHandler(h)
        self._stream_handler = h

    def _close_stream_handlers(self):
        for h in self.__stream_handlers():
            h.close()
            self.removeHandler(h)

    def _close_file_handlers(self):
        for h in self.__file_handlers():
            h.close()
            self.removeHandler(h)

    def __stream_handlers(self):
        for h in self.handlers:
            if type(h) == StreamHandler:
                yield h

    def __file_handlers(self):
        for h in self.handlers:
            if isinstance(h, FileHandler):
                yield h

    def exception_handler(self, e_type, e_msg, e_trace):
        self.err_msg = str(e_msg)

        has_child = not bool(self.child_desig)
        child_frame = e_trace
        s_frame     = e_trace

        # naviagate to most recent stack frame
        while s_frame.tb_next is not None:
            if has_child is False:
                if self.child_desig in _frame_filename(s_frame):
                    child_frame = s_frame
                    has_child = True

            s_frame = s_frame.tb_next

        code_file = _frame_filename(s_frame)
        file = os.path.split(code_file)[1]
        line = s_frame.tb_lineno

        log_msg = '''\n\n\n
        ____________________________   vengeance   ____________________________
              the result 'w+resign' was added to the game information
              
              "{e_msg}"
              error type:   <{e_type}>
              file:   {file}, line: {line} 
        ____________________________   vengeance   ____________________________
        \n\n\n'''.format(name=self.name,
                         e_msg=e_msg,
                         e_type=e_type.__name__,
                         file=file,
                         line=line)

        log_msg = dedent(log_msg)

        # propagate error through base class exception
        self.error(log_msg, exc_info=(e_type, e_msg, child_frame))

        self._close_file_handlers()
        self.callback()

    # def __repr__(self):
    #     return 'log  {}'.format(self.name)


class pool_executor_log_cls(ProcessPoolExecutor):
    def __init__(self, max_workers=None,
                       base_name='pool_executor_log_cls',
                       f_dir=None):

        super().__init__(max_workers)

        self.base_name = base_name
        self.f_dir     = f_dir

    def submit(self, fn, *args, **kwargs):
        kwargs['i'] = self._queue_count
        kwargs['base_name'] = self.base_name
        kwargs['f_dir']     = self.f_dir

        return super().submit(_function_wrapper, fn, *args, **kwargs)


# noinspection PyBroadException
def _function_wrapper(fn, *args, **kwargs):
    i = kwargs.pop('i') + 1
    base_name = kwargs.pop('base_name')
    f_dir     = kwargs.pop('f_dir')

    try:
        return fn(*args, **kwargs)
    except Exception:
        name = '{}_{}.log'.format(base_name, i)
        log_ = log_cls(name, f_dir)
        log_.exception_handler(*sys.exc_info())


def _frame_filename(s_frame):
    return s_frame.tb_frame.f_code.co_filename




# Excel's integer codes for cell errors
excel_errors = {-2146826281: 'error div0',
                -2146826246: 'error na',
                -2146826259: 'error name',
                -2146826288: 'error null',
                -2146826252: 'error num',
                -2146826265: 'error ref',
                -2146826273: 'error value',
                -2146826281: 'error div0'}

# colors
xlClear = -4142

# find parameters
xlCellTypeFormulas = -4123
xlValues           = -4163
xlErrors           = 16
xlWhole            = 1
xlPart             = 2
xlNext             = 1
xlPrevious         = 2
xlByRows           = 1
xlByColumns        = 2
xlUp               = -4162
xlToRight          = -4161

# other enums
xlPasteColumnWidths    = 8
xlMaximized            = -4137
xlCalculationManual    = -4135
xlCalculationAutomatic = -4105






import re
import json

from timeit import default_timer
from datetime import date
from collections import Iterable


def print_runtime(f):
    def runtime_wrapper(*args, **kwargs):
        tic = default_timer() * 1000
        result = f(*args, **kwargs)
        toc = default_timer() * 1000

        elapsed = -(tic - toc)
        print_unicode('\tτ: @{}: {}'.format(function_name(f), format_ms(elapsed)))

        return result

    return runtime_wrapper


def print_performance(f=None, *, repeat=3):
    """
    stolen and modified from
    https://github.com/realpython/materials/blob/master/pandas-fast-flexible-intuitive/tutorial/timer.py
    """
    import gc
    import functools
    import itertools

    functools_repeat = functools.partial(itertools.repeat, None)

    def performance_wrapper(_f_):

        @functools.wraps(_f_)
        def functools_wrapper(*args, **kwargs):
            gcold = gc.isenabled()
            gc.disable()

            result = None
            best   = None
            total  = 0

            try:
                for _ in functools_repeat(repeat):
                    tic = default_timer() * 1000
                    result = _f_(*args, **kwargs)
                    toc = default_timer() * 1000

                    elapsed = -(tic - toc)
                    if best is None:
                        best = elapsed
                    else:
                        best = min(best, elapsed)

                    total += elapsed

                average = total / repeat
                print_unicode('\tτ: @{}\n\t\taverage:   {}\n\t\tbest:      {}\n'
                              .format(function_name(_f_), format_ms(average), format_ms(best)))
            finally:
                if gcold:
                    gc.enable()

            return result
        return functools_wrapper

    if f is None:
        return performance_wrapper
    else:
        return performance_wrapper(f)


def format_ms(ms):
    m, s = divmod(ms / 1000, 60)
    h, m = divmod(m, 60)

    if h >= 1:
        f_ms = '{} hr {} min'.format(h, m)
    elif m >= 1:
        f_ms = '{} min {} sec'.format(int(m), int(s))
    elif s > 1:
        f_ms = '{:.2f} sec'.format(s)
    elif ms > 10:
        f_ms = '{} ms'.format(int(ms))
    elif ms > 1:
        f_ms = '{:.1f} ms'.format(ms)
    else:
        f_ms = '{:.2f} us'.format(int(ms * 1000))

    return f_ms


def function_name(f):
    return '{}.{}'.format(f.__module__.split('.')[-1], f.__name__)


def repr_(sequence, concat=', ', quotes=False, wrap=True):
    """ extend formatting options beyond built-in repr() """

    def repr_recurse(v):
        if isinstance(v, str):
            if quotes:
                v = "'{}'".format(v)

            return v

        if is_sequence_dict:
            c = ': '
            w = False

            return repr_(v, c, quotes, w)

        if isinstance(v, Iterable):
            c = ', '
            w = wrap

            return repr_(v, c, quotes, w)

        return str(v)

    def wrapped_characters(v):
        if isinstance(v, dict):
            w = '{}'
        elif isinstance(v, tuple):
            w = '()'
        elif isinstance(v, list):
            w = '[]'
        else:
            w = None

        return w

    if wrap is True:
        w_ = wrapped_characters(sequence)
    else:
        w_ = wrap

    is_sequence_dict = isinstance(sequence, dict)
    if is_sequence_dict:
        sequence = sequence.items()

    s = concat.join(repr_recurse(o) for o in sequence)

    if w_:
        s = w_[0] + s + w_[1]

    return s


def sanitize(v):
    """
    convert value to string and remove common characters that can make
    string more difficult to work with (quotes and whitespace)

    (the html non-breaking space (&nbsp) is particularly annoying, but dealt
    with by the regex whitespace character class)
    """
    non_standard_re = re.compile(r'[\"\'\s]+')

    s = value_to_string(v)
    s = non_standard_re.sub('', s).strip()

    return s


def value_to_string(v):
    if v is None:
        return ''

    return str(v)


def remove_multiple(s, substrs, first_only=False):
    if s == '':
        return s

    for sub in substrs:
        if first_only:
            s = s.replace(sub, '', 1)
        else:
            s = s.replace(sub, '')

    return s


def between(s, substr_1, substr_2=None):
    """ text between inner and outer substrings (exclusive)
    eg:
        'ell' = between('hello', 'h', 'o')
    """
    if substr_2 is None:
        substr_2 = substr_1

    i_1 = s.index(substr_1) + len(substr_1)
    i_2 = s.index(substr_2, i_1)

    return s[i_1:i_2]


def p_json_dumps(o, indent=4, ensure_ascii=False):
    """
    json can not convert certain python objects to string representations
        dates
        sets
    """

    def unhandled_conversion(_o):
        if isinstance(_o, date):
            return _o.isoformat()

        if isinstance(_o, set):
            return list(_o)

        raise TypeError('cannot convert type to json ' + repr(_o))

    s = json.dumps(o, indent=indent, default=unhandled_conversion, ensure_ascii=ensure_ascii)

    return s


def to_ascii(s):
    greek_chrs = {'α': 'a',
                  'β': 'b',
                  'ε': 'e',
                  'τ': 't',
                  'ν': 'v'}

    s = str(s)
    for old, new in greek_chrs.items():
        s = s.replace(old, new)

    s = (s.encode('ascii', errors='backslashreplace')
         .decode('ascii'))

    return s


def print_unicode(s):
    try:
        print(s)
    except UnicodeEncodeError:
        print(to_ascii(s))


def vengeance_message(s):
    print_unicode('\tν: {}'.format(s))




# noinspection PyUnresolvedReferences
from pythoncom import com_error

from datetime import date
from datetime import datetime

from .. util.iter import modify_iteration_depth

from .. excel_com.excel_address import col_letter
from .. excel_com.excel_address import col_number
from .. excel_com.excel_address import max_cols as excel_max_cols
from .. excel_com.excel_address import max_rows as excel_max_rows

from . workbook import app_to_foreground
from . excel_constants import *


def get_worksheet(wb,
                  ws,
                  *,
                  clear_filter=False,
                  activate=False):

    if ws.__class__.__name__ in {'CDispatch', '_Worksheet'}:
        return ws

    if wb is None:
        raise AssertionError("Excel workbook has not been set, cannot retrieve worksheet: '{}'".format(ws))

    try:
        ws = wb.Sheets[ws]
    except com_error as e:
        raise NameError("'{}' worksheet not found in '{}'".format(ws, wb.Name)) from e

    if clear_filter:
        clear_worksheet_filter(ws)

    if activate:
        ws.Visible = True
        app_to_foreground(ws.Application)
        ws.Activate()

    return ws


def first_row(excel_range, default_r=1):
    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByRows,
                        search_direction=xlNext)

    if search:
        r = max(search.Row, default_r)
    else:
        r = default_r

    return r


def last_row(excel_range, default_r=1):
    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByRows,
                        search_direction=xlPrevious)

    if search:
        r = max(search.Row, default_r)
    else:
        r = default_r

    return r


def first_col(excel_range, default_c='A'):

    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByColumns,
                        search_direction=xlNext)

    if search:
        c = max(search.Column, col_number(default_c))
        c_lttr = col_letter(c)
    else:
        c_lttr = default_c

    return c_lttr


def last_col(excel_range, default_c='A'):

    search = range_find(excel_range,
                        what='*',
                        look_at=xlPart,
                        search_order=xlByColumns,
                        search_direction=xlPrevious)

    if search:
        c = max(search.Column, col_number(default_c))
        c_lttr = col_letter(c)
    else:
        c_lttr = default_c

    return c_lttr


def range_find(excel_range,
               what,
               after=None,
               look_in=xlValues,
               look_at=xlWhole,
               search_order=xlByRows,
               search_direction=xlPrevious,
               match_case=False):

    if after is None:
        if search_direction == xlNext:
            after = last_cell(excel_range)
        elif search_direction == xlPrevious:
            after = first_cell(excel_range)

    search = excel_range.Find(what,
                              after,
                              look_in,
                              look_at,
                              search_order,
                              search_direction,
                              match_case)

    return search


def first_cell(excel_range):
    return excel_range.Cells(1)


def last_cell(excel_range):
    """
    excel_range.Cells(excel_range.Cells.Count) is not always reliable
    and can cause an overflow error, but when excel_range.Address
    only provides row address, something like '$1:$20', this method must be used

    # '$1:$20' = excel_range.Address
    # excel_range.Parent.Rows(a)
    """
    try:
        a = excel_range.Address.split(':')[-1]
        return excel_range.Parent.Range(a)
    except com_error:
        pass

    return excel_range.Cells(excel_range.Cells.Count)


def is_filtered(ws):
    return (ws.AutoFilter is not None) and bool(ws.AutoFilter.FilterMode)


# noinspection PyProtectedMember
def clear_worksheet_filter(ws):
    if is_filtered(ws):
        ws._AutoFilter.ShowAllData()


def is_range_empty(excel_range):
    num_blank = excel_range.Application.WorksheetFunction.CountBlank(excel_range)
    num_cells = excel_range.Cells.Count

    return num_blank == num_cells


def activate_sheet(ws):
    app_to_foreground(ws.Application)
    ws.Visible = True
    ws.Activate()


def write_to_excel_range(v, excel_range):
    m = modify_iteration_depth(v, 2)
    m = list(__excel_friendly_matrix(m))

    a = excel_range.Address
    if ':' not in a:
        a = '{}:{}'.format(a, excel_range.Resize(len(m), len(m[0])).Address)
    else:
        a = excel_range.Resize(len(m), len(m[0])).Address

    excel_range.Parent.Range(a).Value = m


def gen_range_rows(excel_range):
    for row in __convert_excel_errors(excel_range):
        yield row


def __convert_excel_errors(excel_range):
    m = excel_range.Value
    m = modify_iteration_depth(m, 2)
    m = [list(row) for row in m]

    try:
        range_errors = excel_range.SpecialCells(xlCellTypeFormulas, xlErrors)
        cell = first_cell(excel_range)
        r_0  = cell.Row
        c_0  = cell.Column

        for cell_error in range_errors:
            r = cell_error.Row - r_0
            c = cell_error.Column - c_0
            m[r][c] = excel_errors.get(m[r][c], 'unknown error')

    except com_error:
        pass

    return m


def __excel_friendly_matrix(m):
    """ modify matrix values so they can be written to an Excel range without error
    TODO:
        profile / optimize function
        convert date to datetime?
    """
    num_cols = len(m[0])
    num_rows = len(m)

    if num_cols > excel_max_cols:
        raise ValueError("number of columns ({:,}) exceeds Excel's column limit\n"
                         "(did you mean to transpose this matrix?)".format(num_cols))

    if num_rows > excel_max_rows:
        raise ValueError("number of rows ({:,}) exceeds Excel's row limit".format(num_rows))

    for r, row in enumerate(m):
        if len(row) != num_cols:
            _m_ = '\n\t'.join([repr(_row_) for _row_ in m[r - 1:r + 2]])
            raise ValueError('cannot write to Excel, jagged column in matrix'
                             '\nrow {:,}\n\n\t{}'.format(r, _m_))

        _row_ = []
        for c, v in enumerate(row):
            if (v is None) or isinstance(v, (bool, int, float, str)):
                pass
            elif isinstance(v, date):
                v = datetime(v.year, v.month, v.day)
            else:
                v = repr(v)

            _row_.append(v)

        yield _row_


def parse_range(excel_range):
    c_1 = first_cell(excel_range).Column
    c_1 = col_letter(c_1)

    c_2 = last_cell(excel_range).Column
    c_2 = col_letter(c_2)

    r_1 = first_cell(excel_range).Row
    r_2 = last_cell(excel_range).Row

    return c_1, r_1, c_2, r_2





"""
    ./files/example.xlsm, 'invoke python' tab
"""

import os
import sys

from datetime import datetime
from time import sleep

from argparse import ArgumentParser


def parse_cmd_line(fake_cli_str=None):
    print('invoked_from_excel.py')

    if fake_cli_str:
        __add_sys_args(fake_cli_str)

    parser = ArgumentParser()
    parser.add_argument('command')
    parser.add_argument('--content', default='')

    cli_args = parser.parse_args()

    if cli_args.command == 'write_file':
        write_file(cli_args.content)
    elif cli_args.command == 'write_file_and_wait':
        write_file_and_wait(cli_args.content)
    else:
        raise ValueError('invalid command {}'.format(cli_args.command))


def write_file(content):
    f_dir, _ = os.path.split(os.path.realpath(__file__))
    path = f_dir + '\\files\\python_api.txt'

    with open(path, 'w') as f:
        f.write('python script: ' + datetime.now().strftime('%Y-%m-%d %I:%M:%S %p') + ' ' + content)


def write_file_and_wait(content):
    write_file(content)
    print('sleeping ... ')
    sleep(5)


def __add_sys_args(cli_str):
    sys.argv = ['']

    s = cli_str.replace('\n', '')
    for arg in s.split(' ', 2):
        sys.argv.append(arg)


# for debugging in IDE
# parse_cmd_line('write_file_and_wait --content invoke_python_test_1')

parse_cmd_line()




import re

# noinspection PyUnresolvedReferences
from pythoncom import com_error

from collections import namedtuple
from collections import OrderedDict

from .. excel_com.worksheet import activate_sheet
from .. excel_com.worksheet import first_col
from .. excel_com.worksheet import last_col
from .. excel_com.worksheet import first_row
from .. excel_com.worksheet import last_row
from .. excel_com.worksheet import gen_range_rows
from .. excel_com.worksheet import clear_worksheet_filter
from .. excel_com.worksheet import is_filtered
from .. excel_com.worksheet import write_to_excel_range
from .. excel_com.worksheet import is_range_empty
from .. excel_com.worksheet import parse_range

from .. excel_com.excel_address import col_letter
from .. excel_com.excel_address import col_letter_offset
from .. excel_com.excel_address import col_number

from .. excel_com.excel_constants import *

from .. util.iter import index_sequence
from .. util.iter import modify_iteration_depth
from .. util.text import between

from . flux_row_cls import lev_row_cls


class excel_levity_cls:

    allow_focus = False
    
    @staticmethod
    def col_letter_offset(col_str, offset):
        return col_letter_offset(col_str, offset)
    
    @staticmethod
    def col_letter(col_int):
        return col_letter(col_int)

    @staticmethod
    def col_number(col_str):
        return col_number(col_str)

    def __init__(self, ws,
                       *,
                       first_c=None,
                       last_c=None,
                       meta_r=0,
                       header_r=0,
                       first_r=0,
                       last_r=0):

        self.ws      = ws
        self.ws_name = ws.Name

        self.headers   = OrderedDict()
        self.m_headers = OrderedDict()

        self._named_ranges  = _named_ranges_in_workbook(self.workbook)
        self._fixed_columns = first_c, last_c
        self._fixed_rows    = first_r, last_r

        self.first_c  = first_c
        self.last_c   = last_c
        self.meta_r   = meta_r
        self.header_r = header_r
        self.first_r  = first_r
        self.last_r   = last_r

        self.num_cols = 0
        self.num_rows = 0
        self.is_empty = None

        self.set_range_boundaries(index_meta=True, index_header=True)

    @property
    def application(self):
        return self.ws.Application

    @property
    def workbook(self):
        return self.ws.Parent

    @property
    def named_ranges(self):
        return {name: self.excel_range(name) for name in self._named_ranges.keys()}

    @property
    def worksheet(self):
        return self.ws

    @property
    def worksheet_name(self):
        return self.ws_name

    @property
    def meta_headers(self):
        return self.m_headers

    @property
    def header_values(self):
        return list(self.headers.keys())

    @property
    def meta_header_values(self):
        return list(self.m_headers.keys())

    @property
    def has_headers(self):
        if self.is_empty:
            return False

        return bool(self.headers) and bool(self.m_headers)

    @property
    def has_filter(self):
        return bool(self.ws.AutoFilter)

    @property
    def append_r(self):
        """ determines the first available empty row in sheet """
        if self.is_empty:
            r = self.first_r
        else:
            r = self.last_r + 1

        return r

    @property
    def append_c(self):
        """ determines the first available empty column in sheet """
        if self.is_empty:
            c = self.first_c
        else:
            c = col_letter_offset(self.last_c, 1)

        return c

    def activate(self):
        if self.allow_focus:
            activate_sheet(self.ws)

    def clear_filter(self):
        if not self.has_filter:
            return

        if is_filtered(self.ws):
            self.set_range_boundaries()

    def remove_filter(self):
        if not self.has_filter:
            return

        was_filtered = is_filtered(self.ws)
        self.ws.AutoFilterMode = False

        if was_filtered:
            self.set_range_boundaries()

    def reapply_filter(self, c_1='*f', c_2='*l'):
        if self.header_r > 0:
            r = '*h'
        else:
            r = '*f'

        a = '{} {}:{} {}'.format(c_1, r, c_2, r)
        excel_range = self.excel_range(a)

        if excel_range.Cells.Count == 1:
            a = '{} {}:{} {}'.format(c_1, r, c_2, self.last_r + 1)
            excel_range = self.excel_range(a)

        self.remove_filter()
        excel_range.AutoFilter(*(1,))

    def calculate(self):
        self.excel_range('*f *h: *l *l').Calculate()

    def clear(self, reference,
                    clear_values=True,
                    clear_colors=False):

        excel_range = self.excel_range(reference)
        _, r_1, _, r_2 = parse_range(excel_range)

        if clear_values:
            excel_range.ClearContents()

            index_meta   = (r_1 <= self.meta_r)
            index_header = (r_1 <= self.header_r)
            self.set_range_boundaries(index_meta, index_header)

        if clear_colors:
            excel_range.Interior.Color = xlClear

    def set_range_boundaries(self, index_meta=True, index_header=True):
        """ find the edges of data in worksheet

        worksheet filter MUST be cleared from worksheet to correctly determine these boundaries
        """
        clear_worksheet_filter(self.ws)

        self.__find_range_boundaries()

        if index_meta:
            self.__index_meta_columns()

        if index_header:
            self.__index_header_columns()

    def __find_range_boundaries(self):
        self.first_c, self.last_c = self._fixed_columns
        self.first_r, self.last_r = self._fixed_rows

        excel_range = self.ws.UsedRange

        self.first_c = self.first_c or first_col(excel_range)
        self.first_c = col_letter(self.first_c)
        self.last_c  = self.last_c or last_col(excel_range, default_c=self.first_c)
        self.last_c  = col_letter(self.last_c)

        self.num_cols = col_number(self.last_c) - col_number(self.first_c) + 1

        r_1 = max(self.meta_r, self.header_r) + 1
        r_2 = excel_range.Rows.Count
        a = '{}{}:{}{}'.format(self.first_c, r_1, self.last_c, r_2)
        excel_range = self.ws.Range(a)

        self.first_r = self.first_r or first_row(excel_range, default_r=r_1)
        self.last_r  = self.last_r or last_row(excel_range, default_r=self.first_r)

        self.num_rows = self.last_r - self.first_r + 1

        if self.last_r > self.first_r:
            self.is_empty = False
        else:
            r = self.header_r or self.meta_r or self.first_r
            a = '{}{}:{}{}'.format(self.first_c, r, self.last_c, r)
            self.is_empty = is_range_empty(self.ws.Range(a))

    @classmethod
    def index_headers(cls, ws, row_int=None):
        if row_int is None:
            row_int = first_row(ws)

        c = col_letter(ws.UsedRange.Columns.Count)
        a = '{}{}:{}{}'.format('A', row_int, c, row_int)
        excel_range = ws.Range(a)

        return cls.__index_row_headers(excel_range)

    def __index_headers(self, row_ref):
        a = '*f {} :*l {}'.format(row_ref, row_ref)
        excel_range = self.excel_range(a)

        return self.__index_row_headers(excel_range)

    @classmethod
    def __index_row_headers(cls, excel_range):
        row = next(gen_range_rows(excel_range))

        start = excel_range.Column
        headers = index_sequence(row, start, as_strings=True)
        headers = OrderedDict((h, col_letter(v)) for h, v in headers.items())

        return headers

    def __index_meta_columns(self):
        if self.meta_r == 0:
            return

        self.m_headers = self.__index_headers('meta_r')

    def __index_header_columns(self):
        if self.header_r == 0:
            return

        self.headers = self.__index_headers('header_r')

    def excel_range(self, reference):
        a = None
        excel_range = None

        try:
            if reference not in self._named_ranges:
                a = self.excel_address(reference)
                excel_range = self.ws.Range(a)
            else:
                named_range = self._named_ranges[reference]
                a = '{}!{}'.format(named_range.tab_name, named_range.address)
                excel_range = self.workbook.Sheets(named_range.tab_name).Range(named_range.address)
        except com_error:
            pass

        if excel_range is None:
            a = (str(a).replace('NoneNone', 'None')
                       .replace('None', ' None ')
                       .strip())
            raise ValueError("Range reference '{}' resolved to an invalid Excel address: '{}'"
                             .format(reference, a))

        return excel_range

    def excel_address(self, reference):
        if ':' in reference:
            a_1, a_2 = reference.split(':')
            c_1, r_1 = _reference_to_col_row(self, a_1)
            c_2, r_2 = _reference_to_col_row(self, a_2)

            a = '{}{}:{}{}'.format(c_1, r_1, c_2, r_2)
        else:
            c_1, r_1 = _reference_to_col_row(self, reference)
            a = '{}{}'.format(c_1, r_1)

        return a

    def rows(self, r_1='*h', r_2='*l'):
        if self.is_empty:
            return [[]]

        a = '*f {}:*l {}'.format(r_1, r_2)
        excel_range = self.excel_range(a)

        return (row for row in gen_range_rows(excel_range))

    def flux_rows(self, r_1='*h', r_2='*l'):
        if self.is_empty:
            return [[]]

        a = '*f {}:*l {}'.format(r_1, r_2)
        excel_range = self.excel_range(a)

        if self.headers:
            headers = index_sequence(self.headers.keys(), as_strings=True)
        elif self.m_headers:
            headers = index_sequence(self.m_headers.keys(), as_strings=True)
        else:
            headers = {}

        reserved = headers.keys() & lev_row_cls.class_names
        if reserved:
            raise NameError("reserved name(s) {} found in header row {}".format(list(reserved), list(headers.keys())))

        r_1 = excel_range.Row
        c_1, c_2 = self.first_c, self.last_c

        for r, row in enumerate(gen_range_rows(excel_range), r_1):
            a = '${}${}:${}${}'.format(c_1, r, c_2, r)
            yield lev_row_cls(headers, row, a)

    def __validate_destination_size(self, m, c_1, r_1):
        """
        if lev has fixed columns or rows, these should not be exceeded
        make sure matrix fits in allowed destination space
        """
        num_cols = len(m[0])
        num_rows = len(m)

        first_c, last_c = self._fixed_columns
        first_r, last_r = self._fixed_rows

        if last_c:
            first_c = col_number(first_c) or col_number(c_1)
            last_c  = col_number(last_c)
            col_max = (last_c - first_c) + 1

            if num_cols > col_max:
                raise ValueError('Number of columns in data exceeds fixed destination range')

        if last_r:
            first_r = int(first_r) or int(r_1)
            last_r  = int(last_r)
            row_max = (last_r - first_r) + 1

            if num_rows > row_max:
                raise ValueError('Number of rows in data exceeds fixed destination range')

    def __getitem__(self, reference):
        return self.excel_range(reference)

    def __setitem__(self, reference, v):
        """ write value(s) to excel range """
        excel_range = self.excel_range(reference)
        m = modify_iteration_depth(v, 2)

        c = excel_range.Column
        r = excel_range.Row
        self.__validate_destination_size(m, c, r)

        write_to_excel_range(m, excel_range)

        index_meta   = (r <= self.meta_r)
        index_header = (r <= self.header_r)
        self.set_range_boundaries(index_meta, index_header)

        if self.has_filter:
            self.reapply_filter()

    def __iter__(self):
        for row in self.flux_rows('*f'):
            yield row

    def __repr__(self):
        if self.first_c and self.last_c:
            a = "{}{}:{}{}".format(self.first_c,
                                   self.header_r,
                                   self.last_c,
                                   self.last_r)
        else:
            a = '{unknown address}'

        return "'{}' {}".format(self.ws_name, a)


def _named_ranges_in_workbook(wb):
    named_ranges = {}
    named_range_ntc = namedtuple('named_range_ntc', ('tab_name', 'address'))

    for nr in wb.Names:
        name = nr.Name.lower()
        full_addr = nr.RefersTo.lower()

        if ('!' not in name
          and '_xl' not in name
          and '#ref' not in full_addr):

            a = full_addr.split('\\')[-1]
            a = a.replace("'", '')
            a = a.replace('=', '')

            if '[' in a and ']' in a:
                wb_name = '[' + between(a, '[', ']') + ']'          # [workbook.xlsx]tab_name!address
                a = a.replace(wb_name, '')

            tab_name, a = a.split('!')
            named_ranges[nr.Name] = named_range_ntc(tab_name, a.upper())

    return named_ranges


def _reference_to_col_row(lev, reference):
    reference = __reference_to_property_names(reference)
    col, row  = __reference_to_property_values(lev, reference)
    col, row  = __parse_characters_from_digits(reference, col, row)

    return col, row


def __reference_to_property_names(reference):
    """
    eg:
        'header_c first_r' = __reference_to_property_names('*h *f')
    """
    anchor_names = {'*m': 'meta',
                    '*h': 'header',
                    '*f': 'first',
                    '*l': 'last',
                    '*a': 'append'}

    anchor_re = re.compile('''
         (?P<col>^[*][fla])
        |(?P<row>[*][mhfla]$)
    ''', re.X | re.I)

    reference = reference.strip()

    for match in anchor_re.finditer(reference):
        name  = match.lastgroup
        value = match.group(0)

        if name == 'col':
            col = anchor_names[value] + '_c '
            reference = reference.replace(value, col, 1)
        elif name == 'row':
            row = ' ' + anchor_names[value] + '_r'
            reference = reference.replace(value, row, 1)

    # replace multiple spaces with single space
    reference = ' '.join(reference.split())

    return reference


def __reference_to_property_values(lev, reference):
    if ' ' not in reference:
        return None, None

    # replace multiple spaces with single space
    reference = ' '.join(reference.split())

    splits = reference.split(' ')
    col = __lookup_col_row_on_object(lev, splits[0])
    row = __lookup_col_row_on_object(lev, splits[1])

    return col, row


def __lookup_col_row_on_object(lev, reference):
    if reference in lev.headers:
        literal = lev.headers[reference]
    elif reference in lev.m_headers:
        literal = lev.m_headers[reference]
    elif reference in lev.__dict__:
        literal = lev.__dict__[reference]
    elif reference in lev.__class__.__dict__:
        literal = getattr(lev, reference)
    else:
        literal = None

    return literal


def __parse_characters_from_digits(reference, col, row):
    if (col is not None) and (row is not None):
        return col, row

    address_re = re.compile(r'''
         (?P<col>^[$]?[a-z]{1,2})(?=[\d* ])
        |(?P<row>[$]?[\d]+$)
    ''', re.X | re.I)

    for match in address_re.finditer(reference):
        name  = match.lastgroup
        value = match.group(0)

        if (name == 'col') and (col is None):
            col = value
        elif (name == 'row') and (row is None):
            row = value

    return col, row

